# -*- coding: utf-8 -*-
import os
from gos.exceptions import GOSTaskException, GOSIOException
from gos.utils.load import Loader


class BaseTask(object):
    name = "BaseTask"
    self_loop = False
    do_self_loop = False

    def run(self, manager):
        raise NotImplemented("run method shall be implemented for all the subclasses of BaseTask")


class TaskLoader(object):

    def load_tasks_from_file(self, file_path):
        """ Imports specified python module and returns subclasses of BaseTask from it

        :param file_path: a fully qualified file path for a python module to import CustomTasks from
        :type file_path: `str`
        :return: a dict of CustomTasks, where key is CustomTask.name, and value is a CustomClass task itself
        :rtype: `dict`
        """
        file_name, module_path, objects = Loader.import_custom_python_file(file_path)
        result = {}
        for entry in objects:
            try:
                if issubclass(entry, BaseTask):
                    if entry.__name__ != BaseTask.__name__ and entry.name == BaseTask.name:
                        raise GOSTaskException("Class {class_name} form file {file_name} does not have a unique `name` class field. "
                                               "All custom tasks must have a unique `name` class field for them, tat is used for future reference"
                                               "".format(class_name=entry.name, file_name=os.path.join(module_path, file_name)))
                    result[entry.name] = entry
            except TypeError:
                continue
        return result

    def load_tasks_from_dir(self, dir_path, propagate_exceptions=False):
        """ Imports all python modules in specified directories and returns subclasses of BaseTask from them

        :param propagate_exceptions: a flag that indicates if exceptions from single file import shall be raised during the
            whole directory lookup
        :param dir_path: fully qualified directory path, where all python modules will be search for subclasses of BaseTask
        :type dir_path: `str`
        :return: a dict of CustomTasks, where key is CustomTask.name, and value is a CustomClass task itself
        :rtype: `dict`
        """
        if not os.path.exists(dir_path):
            raise GOSTaskException()
        if os.path.isfile(dir_path):
            raise GOSTaskException()
        result = {}
        for file_basename in os.listdir(path=dir_path):
            full_file_path = os.path.join(dir_path, file_basename)
            try:
                result.update(self.load_tasks_from_file(full_file_path))
            except (GOSTaskException, GOSIOException):
                if propagate_exceptions:
                    raise
        return result

    def load_tasks(self, paths, propagate_exception=False):
        """ Loads all subclasses of BaseTask from modules that are contained in supplied directory paths or direct module paths

        :param propagate_exception: a flag that indicates if exceptions from single file import shall be raised during the
            whole directory lookup
        :param paths: an iterable of fully qualified paths to python modules / directories, from where we import subclasses of BaseClass
        :type paths: `iterable`(`str`)
        :return: a dict of CustomTasks, where key is CustomTask.name, and value is a CustomClass task itself
        :rtype: `dict`
        """
        try:
            result = {}
            for path in paths:
                try:
                    if os.path.isdir(path):
                        result.update(self.load_tasks_from_dir(dir_path=path, propagate_exceptions=propagate_exception))
                    elif os.path.isfile(path):
                        result.update(self.load_tasks_from_file(file_path=path))
                except (GOSTaskException, GOSIOException):
                    if propagate_exception:
                        raise
            return result
        except TypeError:
            raise GOSTaskException("Argument for `load_tasks` method must be iterable")



# -*- coding: utf-8 -*-
import os

import sys

__author__ = "aganezov"
__email__ = "aganezov(at)gwu.edu"
__status__ = "development"


class Configuration(dict):
    """
    Structure, not yaml, just similar looking. It is just a description.

    dir: ./                                               # -- directory, that is used further for creation of relative subdirectories
                                                          #    current working directory by default
    logger:                                               #
        name: GOSLogger                                   #
        level: info                                       #  -- by default only info messages are shown
                                                          #
        format: %(asctime)s - %(name)s - %(levelname)s - %(message)s  #
                                                          #
        destination: sys.stdout                           # -- by default all logging messages are shown to the standard output
                                                          #
    io_silent_fail: false                                 # -- by default if any IO exception happens, program will terminate
                                                          ########################################################################
                                                          ########################################################################
    input:                                                #
        source:                                           #
            - path: file1_path                            # -- relative (to input->dir) path to file with genome data
              format: grimm                               # -- determines which reader is going ot process the source file
                                                          #    if not specified, automatically retrieved from file extension
              io_silent_fail: input->io_silent_fail       # -- whether to fail or not if exception has occurred during data reading
                                                          #    file specific. defaults to input io_silent_fail setting.
                                                          ########################################################################
            - path: file2_path                            #
            - path: file3_path                            #
        dir: .->dir + /input                              # -- a directory to search for source files in. / by default
                                                          ########################################################################
        io_silent_fail: .->io_silent_fail                 # -- input section wide setting whether to fail or not, when an exception
                                                          #    has occurred during the source file processing. Can be overwritten
                                                          #    by source file specific variable
                                                          ########################################################################
        logger: .->logger                                 # -- a logger specification to be utilized for the input section of the
                                                          #
        genomes:                                          # -- if specified, a check performed to make sure data about all
                                                          #    genomes if present. If not -- retrieved form source files
            - name: genome1_name                          # -- primary genome name.
                                                          #    must be unique among all observed genomes (specified and retrieved)
              aliases: [alias1, alias2, alias3]           # -- other genome names, to be identified by. must be unique per genome
                                                          ########################################################################
            - name: genome2_name                          #
            - name: genome3_name                          #
                                                          ########################################################################
    algorithm:                                            ########################################################################
        io_silent_fail: .-> io_silent_fail                # -- if any part of an algorithm performs io operation, this flag determines what
                                                          #    to do if an IO exception is thrown
        logger: .->logger                                 # --
        tasks:                                            # -- single processing entity specification
            paths: []                                     # -- unchangeable value is "./tasks". everything else specified will be appended

                                                          #    to "./tasks" directory. All *.py files are observed and all classes,
                                                          #    being subclasses of GOSTask will be processed and available for further usage
                                                          #################################################################################
        executable_containers:
            - name: stage
              reference: stages
              entry_type_name: task

            - name: round
              reference: rounds
              entry_type_name: stage

        stages:                                           # -- section describing next level layer of processing entities "stages"
            - name: stage1                                # -- unique name of a stages, that it can be referenced by later.
              self_loop: false                            # -- flag determining if a stage must be executed again, after its first execution
                                                          #    is finished.
              tasks:                                      # -- ordered list of tasks that stage includes in itself and will execute
                - task1                                   # -- name based reference to previously specified task
                - task2                                   # --
            - name: stage2                                # --
              logger: algorithm->logger                   # -- logger can be specified uniquely for each stage.
              self_loop: true                             # --
              tasks:                                      # --
                - task2                                   # --
            - name: stage3                                # --
              path: path_to_*.py_file                     # -- if "path" value is specified, the stage is loaded from specified .py file and
                                                          #    its structure is retrieved from the class based attributes
                                                          #################################################################################
        rounds:                                           # -- section describing next level layer of processing entities "rounds"
            - name: round1                                # -- unique name of a round, that can be referenced later
              self_loop: false                            # -- flag determining if a round must be executed again, after its first execution
                                                          #    is finished.
              logger: algorithm->logger                   # -- logger can be specified uniquely for each stage.
              stages:                                     # -- ordered list of stages that round includes in itself and will execute
                - stage1                                  # --
                - stage2                                  # --
            - name: round2                                # --
              path: path_to_*.py_file                     # -- if "path" value is specified, the round is loaded from specified .py file and
                                                          #    its structure is retrieved from the class based attributes
                                                          #################################################################################
        pipeline:                                         # -- top level procession entity
            logger: algorithm->logger                     # --
            self_loop: false                              # --
            rounds:                                       # -- ordered list of rounds that pipeline includes in itself and will execute
                - round1                                  # --
                - round2                                  # --
                - round1                                  # --
                                                          ########################################################################
    output:                                               ########################################################################
        dir: .->dir + output/                             # -- directory for all output files to be put. Used for further paths construction
        logger: .->logger                                 # -- logger specification tp be utilized in the output section
        io_silent_fail: .->io_silent_fail                 # -- output section wide setting to fail or not when an exception
                                                          #    has occurred during the source file processing. Can be overwritten
                                                          #    by output section specific variable
                                                          #################################################################################
        stats:                                            # -- output section which handles all the statistics output for current
                                                          #    scaffolder execution
            dir: output->dir + stats/                     # -- directory where statistics files will be located
            file: stats.txt                               # -- default file name for the overall statistics file
            logger: output->logger                        #
            io_silent_fail: output->io_silent_fail        #
                                                          #################################################################################
        assembly_points:                                  # -- output section which handles all the information output about assembly
                                                          #    points during scaffolder run
            dir: output->dir + assembly_points/           # -- directory where assembly points file will be located
            file: assembly_points.txt                     # -- default name for the overall statistic
            logger: output->logger                        #
            io_silent_fail: output->io_silent_fail        #
            genome_specific: true                         # -- when specified, besides the overall file with all assembly points for current
                                                          #    scaffolder run, also "per-genome" files are created, that duplicate genome
                                                          #    specific results
                                                          #################################################################################
            genome_specific_file_name_pattern: assembly_points_{genome_name}.txt    # pattern for genome_specific file name creation
                                                          #################################################################################
        genomes:                                          # -- output section where information about genomes fragments will be stored
            dir: output->dir + genomes/                   # -- directory where all genomes will ba located
            output_non_glued_fragments: false             # -- if specified, all input information about genomes will be outputted,
                                                          #    if set to false, only those fragments, that we involved in at least one
                                                          #    gluing will be present in the output
            logger: output->logger                        #
            io_silent_fail: output->io_silent_fail        #
    """

    DIR = "dir"
    LOGGER = "logger"
    IOSF = "io_silent_fail"
    INPUT = "input"
    ALGORITHM = "algorithm"
    OUTPUT = "output"
    SOURCE = "source"
    NAME = "name"
    LEVEL = "level"
    FORMAT = "format"
    DESTINATION = "destination"
    ASSEMBLY_POINTS = "assembly_points"
    GENOMES = "genome"
    STATS = "stats"
    TASKS = "tasks"
    STAGES = "stages"
    ROUNDS = "rounds"
    PIPELINE = "pipeline"
    PATH = "path"
    PATHS = "paths"
    FILE = "file"
    GENOME_SPECIFIC = "genome_specific"
    GENOME_SPECIFIC_FNP = "genome_specific_file_name_pattern"
    OUTPUT_NG_FRAGMENTS = "output_non_glued_fragments"
    SELF_LOOP = "self_loop"
    EXECUTABLE_CONTAINERS = "executable_containers"
    ENTRIES = "entries"
    REFERENCE = "reference"

    # predefined constants
    DEFAULT_IOSF = False
    DEFAULT_LOGGER_NAME = "GOSLogger"
    DEFAULT_LOGGER_LEVEL = "info"
    DEFAULT_LOGGER_FORMAT = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    DEFAULT_INPUT_DIR = "input"
    DEFAULT_OUTPUT_DIR = "output"
    DEFAULT_OUTPUT_STATS_DIR = "stats"
    DEFAULT_OUTPUT_STATS_FILE = "stats.txt"
    DEFAULT_OUTPUT_AP_FILE = "assembly_points.txt"
    DEFAULT_OUTPUT_AP_DIR = "assembly_points"
    DEFAULT_OUTPUT_AP_GENOME_SPECIFIC = True
    DEFAULT_OUTPUT_AP_GSFNP = "assembly_points_{genome_name}.txt"
    DEFAULT_OUTPUT_GENOMES_ONGF = False
    DEFAULT_OUTPUT_GENOMES_DIR = "genomes"
    DEFAULT_ALGORITHM_TASKS_PATH = "./tasks"
    DEFAULT_ALGORITHM_PIPELINE_SELF_LOOP = True
    DEFAULT_ALGORITHM_EC_SELF_LOOP = True
    DEFAULT_LOGGER_DESTINATION = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._init_top_level_fields()
        self._init_logger_top_level_section()
        self._init_input_section()
        self._init_algorithm_section()
        self._init_output_section()

    def _init_output_section(self):
        if self.OUTPUT not in self:
            self[self.OUTPUT] = {}
        if self.DIR not in self[self.OUTPUT]:
            self[self.OUTPUT][self.DIR] = None
        if self.LOGGER not in self[self.OUTPUT]:
            self[self.OUTPUT][self.LOGGER] = {}
        if self.IOSF not in self[self.OUTPUT]:
            self[self.OUTPUT][self.IOSF] = None
        if self.ASSEMBLY_POINTS not in self[self.OUTPUT]:
            self[self.OUTPUT][self.ASSEMBLY_POINTS] = {}
        if self.GENOMES not in self[self.OUTPUT]:
            self[self.OUTPUT][self.GENOMES] = {}
        if self.STATS not in self[self.OUTPUT]:
            self[self.OUTPUT][self.STATS] = {}

    def _init_algorithm_section(self):
        if self.ALGORITHM not in self:
            self[self.ALGORITHM] = {}
        if self.LOGGER not in self[self.ALGORITHM]:
            self[self.ALGORITHM][self.LOGGER] = {}
        if self.IOSF not in self[self.ALGORITHM]:
            self[self.ALGORITHM][self.IOSF] = None
        if self.EXECUTABLE_CONTAINERS not in self[self.ALGORITHM]:
            self[self.ALGORITHM][self.EXECUTABLE_CONTAINERS] = []
        if self.TASKS not in self[self.ALGORITHM]:
            self[self.ALGORITHM][self.TASKS] = {}
        if self.PIPELINE not in self[self.ALGORITHM]:
            self[self.ALGORITHM][self.PIPELINE] = {}

    def _init_input_section(self):
        if self.INPUT not in self:
            self[self.INPUT] = {}
        if self.LOGGER not in self[self.INPUT]:
            self[self.INPUT][self.LOGGER] = {}
        if self.DIR not in self[self.INPUT]:
            self[self.INPUT][self.DIR] = None
        if self.IOSF not in self[self.INPUT]:
            self[self.INPUT][self.IOSF] = None
        if self.SOURCE not in self[self.INPUT]:
            self[self.INPUT][self.SOURCE] = []

    def _init_logger_top_level_section(self):
        if self.LOGGER not in self:
            self[self.LOGGER] = {}
        if self.NAME not in self[self.LOGGER]:
            self[self.LOGGER][self.NAME] = None
        if self.LEVEL not in self[self.LOGGER]:
            self[self.LOGGER][self.LEVEL] = None
        if self.FORMAT not in self[self.LOGGER]:
            self[self.LOGGER][self.FORMAT] = None
        if self.DESTINATION not in self[self.LOGGER]:
            self[self.LOGGER][self.DESTINATION] = None

    def _init_top_level_fields(self):
        if self.DIR not in self:
            self[self.DIR] = None
        if self.IOSF not in self:
            self[self.IOSF] = None

    def update_with_default_values(self):
        """ Goes through all the configuration fields and predefines empty ones with default values

        Top level:
            `dir` field is predefined with current working directory value, in case of empty string or `None`
            `io_silent_fail` field if predefined with :attr:`Configuration.DEFAULT_IOSF` in case of None or empty string

        Logger section:
            `name` field is predefined with :attr:`Configuration.DEFAULT_LOGGER_NAME`. Field is set to str() of itself
            `level` field is predefined with :attr:`Configuration.DEFAULT_LOGGER_LEVEL`. Field is set to str() of itself
            `format` field is predefined with :attr:`Configuration.DEFAULT_LOGGER_LEVEL`. Field is set to str() of itself
            `destination` field if predefined with

        Input section:
            `dir` field is predefined with a relative path constructed with top level `dir` field and :attr:`Configuration.DEFAULT_INPUT_DIR`
            `source` field is predefined with an empty list
            `io_silent_fail` if predefined with a top level `io_silent_fail`
            `logger` subsection if predefined by a top level `logger` section it substitutes all the missing values in `input` `logger` subsection

        Algorithm section:
            `io_silent_fail` is predefined with a top level `io_silent_fail` value
            `logger` is predefined with top level `logger` configuration
            `tasks` section:
                `paths` is predefined by [:attr:`Configuration.DEFAULT_ALGORITHM_TASKS_PATH`]. If value is supplied,
                        :attr:`Configuration.DEFAULT_ALGORITHM_TASKS_PATH` is prepended  to the supplied list
            `stages` section: (is predefined with [])
                if any values are predefined, such fields as `io_silent_fail`, `logger` are propagated to stages entries
                `self_loop` value is predefined by :attr:`Configuration.DEFAULT_ALGORITHM_STAGES_SELF_LOOP`
            `rounds` section: (is predefined with [])
                if any values are predefined, such fields as `io_silent_fail`, `logger` are propagated to stages entries
                `self_loop` value is predefined by :attr:`Configuration.DEFAULT_ALGORITHM_ROUNDS_SELF_LOOP`
            `pipeline` section:
                `self_loop` if predefined by :attr:`Configuration.DEFAULT_ALGORITHM_PIPELINE_SELF_LOOP`
                `logger` if predefined with an algorithm->logger configuration. All non specified value are propagated respectively
                `rounds` is predefined with []

        Output section:
            `dir` field is predefined with :attr:`Configuration.DEFAULT_OUTPUT_DIR`
            `io_silent_fail` field is predefined with top level `io_silent_fail`
            `logger` subsection is predefined with a top level logger section, which substitutes all of missing values from top level `logger` section
            `stats` section
                `io_silent_fail` field is predefined with `output->io_silent_fail` value
                `dir` is predefined with :attr:`Configuration.DEFAULT_OUTPUT_STATS_DIR`
                `file` is predefined with :attr:`Configuration.DEFAULT_OUTPUT_STATS_FILE`
                `logger` is defaulted (all, or just missing parts) bu `output->logger` section

            `assembly_points` section
                `io_silent_fail` field is predefined with `output->io_silent_fail` value
                `dir` is predefined with :attr:`Configuration.DEFAULT_OUTPUT_AP_DIR`
                `file` is predefined with :attr:`Configuration.DEFAULT_OUTPUT_AP_FILE`
                `logger` is defaulted (all, or just missing parts) bu `output->logger` section
                `genome_specific` is predefined with :attr:`Configuration.DEFAULT_OUTPUT_AP_GENOME_SPECIFIC`
                `genome_specific_file_name_pattern` is predefined with :attr:`Configuration.DEFAULT_OUTPUT_AP_GSFNP`

            `genomes` section
                `io_silent_fail` field is predefined with `output->io_silent_fail` value
                `dir` is predefined with :attr:`Configuration.DEFAULT_OUTPUT_GENOMES_DIR`
                `logger` is defaulted (all, or just missing parts) bu `output->logger` section
                `output_non_glued_fragments` is predefined with :attr:`Configuration.DEFAULT_OUTPUT_GENOMES_ONGF`

        :return: Nothing, performs inplace changes
        :rtype: `None`
        """
        if self[self.DIR] in ("", None):
            self[self.DIR] = os.getcwd()
        if self[self.IOSF] in ("", None):
            self[self.IOSF] = self.DEFAULT_IOSF

        # logger section
        if self[self.LOGGER][self.NAME] in ("", None):
            self[self.LOGGER][self.NAME] = self.DEFAULT_LOGGER_NAME
        self[self.LOGGER][self.NAME] = str(self[self.LOGGER][self.NAME])
        if self[self.LOGGER][self.LEVEL] in ("", None):
            self[self.LOGGER][self.LEVEL] = self.DEFAULT_LOGGER_LEVEL
        self[self.LOGGER][self.LEVEL] = str(self[self.LOGGER][self.LEVEL])
        if self[self.LOGGER][self.FORMAT] in ("", None):
            self[self.LOGGER][self.FORMAT] = self.DEFAULT_LOGGER_FORMAT
        self[self.LOGGER][self.FORMAT] = str(self[self.LOGGER][self.FORMAT])
        if self[self.LOGGER][self.DESTINATION] in ([], "", None):
            self[self.LOGGER][self.DESTINATION] = self.DEFAULT_LOGGER_DESTINATION

        # input section
        if self[self.INPUT][self.SOURCE] in ("", None):
            self[self.INPUT][self.SOURCE] = []
        if self[self.INPUT][self.DIR] in ("", None):
            self[self.INPUT][self.DIR] = self.DEFAULT_INPUT_DIR
        if self[self.INPUT][self.IOSF] in ("", None):
            self[self.INPUT][self.IOSF] = self[self.IOSF]
        self._update_logger_config(logger_to_update=self[self.INPUT][self.LOGGER],
                                   source_logger=self[self.LOGGER])

        # algorithm section
        if self.LOGGER not in self[self.ALGORITHM] or self[self.ALGORITHM][self.LOGGER] in ("", None):
            self[self.ALGORITHM][self.LOGGER] = {}
        self._update_logger_config(logger_to_update=self[self.ALGORITHM][self.LOGGER],
                                   source_logger=self[self.LOGGER])
        if self.IOSF not in self[self.ALGORITHM] or self[self.ALGORITHM][self.IOSF] in ("", None):
            self[self.ALGORITHM][self.IOSF] = self[self.IOSF]
        if self.TASKS not in self[self.ALGORITHM]:
            self[self.ALGORITHM][self.TASKS] = {}
        if self.EXECUTABLE_CONTAINERS not in self[self.ALGORITHM]:
            self[self.ALGORITHM][self.EXECUTABLE_CONTAINERS] = []

        if self.PATHS not in self[self.ALGORITHM][self.TASKS] or self[self.ALGORITHM][self.TASKS][self.PATHS] in ("", None):
            self[self.ALGORITHM][self.TASKS][self.PATHS] = []
        self[self.ALGORITHM][self.TASKS][self.PATHS] = [self.DEFAULT_ALGORITHM_TASKS_PATH] + self[self.ALGORITHM][self.TASKS][self.PATHS]

        for ecs in self[self.ALGORITHM][self.EXECUTABLE_CONTAINERS]:
            if self.REFERENCE not in ecs:
                ecs[self.REFERENCE] = ecs[self.NAME] + "s"
            if ecs[self.REFERENCE] not in self[self.ALGORITHM]:
                self[self.ALGORITHM][ecs[self.REFERENCE]] = []
            for executable_container in self[self.ALGORITHM][ecs[self.REFERENCE]]:
                if self.SELF_LOOP not in executable_container:
                    executable_container[self.SELF_LOOP] = self.DEFAULT_ALGORITHM_EC_SELF_LOOP
                if self.ENTRIES not in executable_container:
                    executable_container[self.ENTRIES] = []

        if self.PIPELINE not in self[self.ALGORITHM]:
            self[self.ALGORITHM][self.PIPELINE] = {}
        if self.LOGGER not in self[self.ALGORITHM][self.PIPELINE] or self[self.ALGORITHM][self.PIPELINE][self.LOGGER] in ("", None):
            self[self.ALGORITHM][self.PIPELINE][self.LOGGER] = {}
        self._update_logger_config(logger_to_update=self[self.ALGORITHM][self.PIPELINE][self.LOGGER],
                                   source_logger=self[self.ALGORITHM][self.LOGGER])
        if self.IOSF not in self[self.ALGORITHM][self.PIPELINE] or self[self.ALGORITHM][self.PIPELINE][self.IOSF] in ("", None):
            self[self.ALGORITHM][self.PIPELINE][self.IOSF] = self[self.ALGORITHM][self.IOSF]
        if self.ENTRIES not in self[self.ALGORITHM][self.PIPELINE] or self[self.ALGORITHM][self.PIPELINE][self.ENTRIES] in ("", None):
            self[self.ALGORITHM][self.PIPELINE][self.ENTRIES] = []
        if self.SELF_LOOP not in self[self.ALGORITHM][self.PIPELINE] or self[self.ALGORITHM][self.PIPELINE][self.SELF_LOOP] in ("", None):
            self[self.ALGORITHM][self.PIPELINE][self.SELF_LOOP] = self.DEFAULT_ALGORITHM_PIPELINE_SELF_LOOP

        # output section
        if self[self.OUTPUT][self.DIR] in ("", None):
            self[self.OUTPUT][self.DIR] = os.path.join(self[self.DIR], self.DEFAULT_OUTPUT_DIR)
        if self[self.OUTPUT][self.IOSF] in ("", None):
            self[self.OUTPUT][self.IOSF] = self[self.IOSF]
        self._update_logger_config(logger_to_update=self[self.OUTPUT][self.LOGGER],
                                   source_logger=self[self.LOGGER])

        # output -> stats section
        if self.DIR not in self[self.OUTPUT][self.STATS] or self[self.OUTPUT][self.STATS][self.DIR] in ("", None):
            self[self.OUTPUT][self.STATS][self.DIR] = self.DEFAULT_OUTPUT_STATS_DIR
        if self.IOSF not in self[self.OUTPUT][self.STATS] or self[self.OUTPUT][self.STATS][self.IOSF] in ("", None):
            self[self.OUTPUT][self.STATS][self.IOSF] = self[self.OUTPUT][self.IOSF]
        if self.FILE not in self[self.OUTPUT][self.STATS] or self[self.OUTPUT][self.STATS][self.FILE] in ("", None):
            self[self.OUTPUT][self.STATS][self.FILE] = self.DEFAULT_OUTPUT_STATS_FILE
        if self.LOGGER not in self[self.OUTPUT][self.STATS]:
            self[self.OUTPUT][self.STATS][self.LOGGER] = {}
        self._update_logger_config(logger_to_update=self[self.OUTPUT][self.STATS][self.LOGGER],
                                   source_logger=self[self.OUTPUT][self.LOGGER])

        # output -> assembly_points section
        if self.DIR not in self[self.OUTPUT][self.ASSEMBLY_POINTS] or self[self.OUTPUT][self.ASSEMBLY_POINTS][self.DIR] in ("", None):
            self[self.OUTPUT][self.ASSEMBLY_POINTS][self.DIR] = self.DEFAULT_OUTPUT_AP_DIR
        if self.IOSF not in self[self.OUTPUT][self.ASSEMBLY_POINTS] or self[self.OUTPUT][self.ASSEMBLY_POINTS][self.IOSF] in ("", None):
            self[self.OUTPUT][self.ASSEMBLY_POINTS][self.IOSF] = self[self.OUTPUT][self.IOSF]
        if self.FILE not in self[self.OUTPUT][self.ASSEMBLY_POINTS] or self[self.OUTPUT][self.ASSEMBLY_POINTS][self.FILE] in ("", None):
            self[self.OUTPUT][self.ASSEMBLY_POINTS][self.FILE] = self.DEFAULT_OUTPUT_AP_FILE
        if self.GENOME_SPECIFIC not in self[self.OUTPUT][self.ASSEMBLY_POINTS] or self[self.OUTPUT][self.ASSEMBLY_POINTS][
            self.GENOME_SPECIFIC]:
            self[self.OUTPUT][self.ASSEMBLY_POINTS][self.GENOME_SPECIFIC] = self.DEFAULT_OUTPUT_AP_GENOME_SPECIFIC
        if self.GENOME_SPECIFIC_FNP not in self[self.OUTPUT][self.ASSEMBLY_POINTS] or self[self.OUTPUT][self.ASSEMBLY_POINTS][
            self.GENOME_SPECIFIC_FNP]:
            self[self.OUTPUT][self.ASSEMBLY_POINTS][self.GENOME_SPECIFIC_FNP] = self.DEFAULT_OUTPUT_AP_GSFNP
        if self.LOGGER not in self[self.OUTPUT][self.ASSEMBLY_POINTS]:
            self[self.OUTPUT][self.ASSEMBLY_POINTS][self.LOGGER] = {}
        self._update_logger_config(logger_to_update=self[self.OUTPUT][self.ASSEMBLY_POINTS][self.LOGGER],
                                   source_logger=self[self.OUTPUT][self.LOGGER])

        # output -> genomes section
        if self.DIR not in self[self.OUTPUT][self.GENOMES] or self[self.OUTPUT][self.GENOMES][self.DIR] in ("", None):
            self[self.OUTPUT][self.GENOMES][self.DIR] = self.DEFAULT_OUTPUT_GENOMES_DIR
        if self.IOSF not in self[self.OUTPUT][self.GENOMES] or self[self.OUTPUT][self.GENOMES][self.IOSF] in ("", None):
            self[self.OUTPUT][self.GENOMES][self.IOSF] = self[self.OUTPUT][self.IOSF]
        if self.OUTPUT_NG_FRAGMENTS not in self[self.OUTPUT][self.GENOMES] or self[self.OUTPUT][self.GENOMES][self.OUTPUT_NG_FRAGMENTS] in (
        "", None):
            self[self.OUTPUT][self.GENOMES][self.OUTPUT_NG_FRAGMENTS] = self.DEFAULT_OUTPUT_GENOMES_ONGF
        if self.LOGGER not in self[self.OUTPUT][self.GENOMES]:
            self[self.OUTPUT][self.GENOMES][self.LOGGER] = {}
        self._update_logger_config(logger_to_update=self[self.OUTPUT][self.GENOMES][self.LOGGER],
                                   source_logger=self[self.OUTPUT][self.LOGGER])

    @staticmethod
    def _update_logger_config(logger_to_update, source_logger):
        if logger_to_update in ("", None):
            logger_to_update = {}
        for key, value in source_logger.items():
            if key not in logger_to_update or logger_to_update[key] in ("", None):
                logger_to_update[key] = source_logger[key]



# -*- coding: utf-8 -*-
import importlib
import os

import sys

from gos.exceptions import GOSIOException


class Loader(object):
    @staticmethod
    def import_custom_python_file(file_path):
        if not os.path.exists(file_path):
            raise GOSIOException("Specified file does not exists")
        if os.path.isdir(file_path):
            raise GOSIOException("Specified path corresponds to a directory, not a file")
        module_path, file_name = os.path.split(file_path)
        if not file_name.endswith((".py", ".pyc")):
            raise GOSIOException("Specified path does not correspond to python file ")
        if module_path not in sys.path:
            sys.path.insert(0, module_path)
        module_name = file_name[:file_name.rfind(".")]
        module = importlib.import_module(module_name)
        objects = [getattr(module, attr_name) for attr_name in dir(module)]
        return file_name, module_path, objects


# -*- coding: utf-8 -*-
import logging
import unittest

import importlib
import tempfile
from gos.algo.executable_containers.base_round import Round
from gos.algo.executable_containers.base_stage import Stage
from gos.algo.executable_containers.pipeline import Pipeline
from gos.exceptions import GOSExecutableContainerException, GOSIOException
from gos.executable_containers import ExecutableContainer
from gos.manager import Manager


class ExecutableContainerTestCase(unittest.TestCase):
    def setUp(self):
        self.executable_container = ExecutableContainer()
        self.ec = self.executable_container
        self.dm = Manager({})
        self.dm.logger = logging.getLogger()

    ################################################################################
    #
    # testing attributes, that are utilized in referencing the EC object itself
    #
    ################################################################################

    def test_name_attribute(self):
        self.assertTrue(hasattr(self.ec, "name"))

    def test_type_name_attribute(self):
        self.assertTrue(hasattr(self.ec, "type_name"))

    def test_group_reference_name_attribute(self):
        self.assertTrue(hasattr(self.ec, "group_reference_name"))

    ################################################################################
    #
    # testing attributes, that are utilized in referencing other EC / tasks that
    #      this EC object is working with
    #
    ################################################################################

    def test_entries_type_name_attribute(self):
        self.assertTrue(hasattr(self.ec, "entries_type_names"))

    def test_entries_info_attribute(self):
        self.assertTrue(hasattr(self.ec, "entries_names"))

    def test_entries_attribute(self):
        self.assertTrue(hasattr(self.ec, "entries"))

    ################################################################################
    #
    # testing internal attributes and methods for the EC object
    #
    ################################################################################
    def test_self_loop_attribute(self):
        self.assertTrue(hasattr(self.ec, "self_loop"))

    def test_do_self_loop_attribute(self):
        self.assertTrue(hasattr(self.ec, "do_self_loop"))

    def test_run_method_existence(self):
        self.assertTrue(hasattr(self.ec, "run"))
        self.assertTrue(callable(getattr(self.ec, "run")))

    def test_logger_attribute(self):
        self.assertTrue(hasattr(self.ec, "logger"))

    ################################################################################
    #
    # testing logic of EC object
    #
    ################################################################################
    def test_default_group_reference_name_attribute(self):
        ec = ExecutableContainer(name="test")
        self.assertEqual(ec.group_reference_name, "tests")

    def test_initialization_with_class_defined_entries_names_value(self):
        entry_ = ["entry1", "entry2", "entry3"]

        class MyEC(ExecutableContainer):
            entries_names = entry_

        value = MyEC()
        self.assertListEqual(value.entries_names, entry_)

    def test_initialization_with_no_defined_entries_names_values(self):
        class MyEC(ExecutableContainer):
            pass

        self.assertListEqual(MyEC().entries_names, [])

    def test_initialization_with_specified_entries_names_value(self):
        class MyEC(ExecutableContainer):
            pass

        value_ = ["task1", "round1", "value1"]
        self.assertListEqual(MyEC(entries_names=value_).entries_names, value_)

    def test_initialization_with_class_defined_entries_type_names(self):
        entry_ = ["task", "round"]

        class MyEC(ExecutableContainer):
            entries_type_names = entry_

        value = MyEC()
        self.assertListEqual(value.entries_type_names, entry_)

    def test_initialization_with_no_defined_entries_type_names_value(self):
        class MyEC(ExecutableContainer):
            pass

        self.assertEqual(MyEC().entries_type_names, [])

    def test_initialization_with_specified_entries_type_names_value(self):
        class MyEC(ExecutableContainer):
            pass

        value_ = ["task", "round", "value"]
        self.assertListEqual(MyEC(entries_type_names=value_).entries_type_names, value_)

    def test_initialization_entries_by_default_are_empty_list(self):
        self.assertListEqual(ExecutableContainer().entries, [])

    def test_setup_from_config_no_name(self):
        with self.assertRaises(GOSExecutableContainerException):
            ExecutableContainer.setup_from_config(manager=self.dm, config={})

    def test_setup_from_config_self_loop_value(self):
        ec = ExecutableContainer.setup_from_config(config={"name": "my_name",
                                                           "self_loop": False},
                                                   manager=self.dm)
        self.assertFalse(ec.self_loop)

    def test_setup_from_config_entries_names(self):
        stage_name_list = ["stage1", "stage2", "stage3"]
        ec = ExecutableContainer.setup_from_config(config={"name": "my_name",
                                                           "entries_names": stage_name_list},
                                                   manager=self.dm)
        self.assertListEqual(ec.entries_names, stage_name_list)

    def test_setup_from_file_file_does_not_exists(self):
        non_existing_path = "non_existing_path.py"
        with self.assertRaises(GOSIOException):
            ExecutableContainer.setup_from_file(non_existing_path, "executable_container")

    def test_setup_from_file_non_python_file(self):
        non_py_file = tempfile.NamedTemporaryFile(mode="wt", suffix=".non_py")
        with self.assertRaises(GOSIOException):
            ExecutableContainer.setup_from_file(non_py_file.name, "executable_container")

    def test_setup_from_file_no_unique_name(self):
        tmp_file = tempfile.NamedTemporaryFile(mode="wt", suffix=".py")
        tmp_file.write(self.get_executable_container_import_string())
        tmp_file.write("""class MyContainer(ExecutableContainer):\n\tdef setup(self):\n\t\tpass""")
        tmp_file.flush()
        importlib.invalidate_caches()
        with self.assertRaises(GOSExecutableContainerException):
            ExecutableContainer.setup_from_file(tmp_file.name, "executable_container")

    def get_executable_container_import_string(self):
        return """from gos.executable_containers import ExecutableContainer\n"""

    def test_setup_from_file_no_setup_method(self):
        tmp_file = tempfile.NamedTemporaryFile(mode="wt", suffix=".py")
        tmp_file.write(self.get_executable_container_import_string())
        tmp_file.write("""class MyContainer(ExecutableContainer):\n\tname="new_executable_container_name" """)
        tmp_file.flush()
        importlib.invalidate_caches()
        with self.assertRaises(GOSExecutableContainerException):
            ExecutableContainer.setup_from_file(tmp_file.name, "executable_container")

    def test_setup_from_file_no_match_by_name_attribute(self):
        tmp_file = tempfile.NamedTemporaryFile(mode="wt", suffix=".py")
        tmp_file.write(self.get_executable_container_import_string())
        tmp_file.write("""class MyContainer(ExecutableContainer):\n\tname="new_ec_name"\n\tdef setup(self):\n\t\tpass""")
        tmp_file.flush()
        importlib.invalidate_caches()
        with self.assertRaises(GOSExecutableContainerException):
            ExecutableContainer.setup_from_file(tmp_file.name, container_name="not_new_ec_name")

    def test_setup_from_file(self):
        tmp_file = tempfile.NamedTemporaryFile(mode="wt", suffix=".py")
        tmp_file.write(self.get_executable_container_import_string())
        tmp_file.write(
                """class MyContainer(ExecutableContainer):\n\tname="my_ec"\n\tdef setup(self):\n\t\tself.entries_names = ["entry1"]\n\t\tself.entries_type_names=["task"] """)
        tmp_file.flush()
        importlib.invalidate_caches()
        result = ExecutableContainer.setup_from_file(tmp_file.name, container_name="my_ec")
        self.assertIsInstance(result, ExecutableContainer)
        self.assertListEqual(result.entries_names, ["entry1"])
        self.assertListEqual(result.entries_type_names, ["task"])


class BaseStageTestCase(unittest.TestCase):
    def test_base_stage_executable_container_entries_type_name_attribute(self):
        self.assertListEqual(Stage.entries_type_names, ["task"])

    def test_base_stage_executable_container_type_name(self):
        self.assertEqual(Stage.type_name, "stage")


class BaseRoundTestCase(unittest.TestCase):
    def test_base_round_executable_container_entries_type_name_attribute(self):
        self.assertListEqual(Round.entries_type_names, ["stage"])

    def test_base_round_executable_container_type_name(self):
        self.assertEqual(Round.type_name, "round")


class PipelineTestCase(unittest.TestCase):
    def test_pipeline_executable_container_entries_type_name_attribute(self):
        self.assertIsNone(Pipeline.entries_type_names)

    def test_pipeline_executable_container_type_name(self):
        self.assertEqual(Pipeline.type_name, "pipeline")


if __name__ == '__main__':
    unittest.main()
